/*
  bluegps.c
  Original
  http://www.kowoma.de/gps/zusatzerklaerungen/NMEA.htm

  Modified fixed, and maybe bugified by Martti@Lumme.fi
  http://www.lumme.fi/box/bluegps
  
  1 endian detection to use it on older Mac's
  2 removing bluetooth, but using it through serial interface linkage
  3 added download again from record, because link drops frames
  4 added checking the last frame reception, until eof accepted
  5 added automatic download filename generation
  6 options to select output format
  7 added tab separated format
  8 open file for writing only after loading starts
  
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <string.h>

#include <netinet/ip.h>
#include <netinet/udp.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/wait.h>
#include <errno.h>
#include <math.h>

#include "options.h"

#include "rbt3000.h"

#define NAUTIC_MILE  (1.85185)
#define KPH2KNOTS(a) ((a)/NAUTIC_MILE)


int bt;
int req_quit = 0;

// parameters given via command line options:
int erase = 0;
char *password = NULL;
int records = 0; // records received

char *LogfileName;

char statonly;

// output format options
#define oGPGGA 1
#define oGPRMC 2
#define	oGPGLL 4
#define	oGPVTG 8
#define	oGPGSA 16
#define oAll 31
#define oTabbed 0

int outFormat; // one of the above or tabbed if no selctions

FILE *export_file = NULL; // filename of export file
int export_num, export_expect;  // follow the flow

char *SerialPort;  // name of the serial device

fd_set rfds, wfds; // 

// send one byte to the channel
unsigned char send_byte(unsigned char byte) {
  write(bt, &byte, 1);
  return byte;
}

// send command sequence with pre and postables and checksum
void send_cmd_sequence(unsigned char *data, int len)
{
unsigned char sum;

debug(5,"Sending:\n%s",dump("Data",data,len));

	send_byte(0xaa); send_byte(0x55);
	send_byte(len & 0xff);  send_byte(len >> 8);
	sum = 0;
	while(len--) {
		sum ^= send_byte(*data++);
	}
	send_byte(sum);
	send_byte(0x33); send_byte(0xcc);
}

// send one byte command
void send_cmd(unsigned char code) {
  send_cmd_sequence(&code, 1);
}

// initialize download
void rbt3000_init(void) {
unsigned char cmd1[] = { 
	0xa0, 0xa2, 0x00, 0x31, 0xa5, 0x00, 0x04, 0x04, 
	0x00, 0x00, 0xe1, 0x00, 0x08, 0x01, 0x00, 0x00, //0000e100=57600bps,8data,1 stop,NP,res
	0x00, 0xff, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00, //other uarts 1,2,3
	0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x05, 0x05,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0xff, 0x05, 0x05, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0xb2, 0xb0,
	0xb3 };

	printf("Initializing RBT-3000 ...\n");

// sending $RCMD402*2E followed by SiRF
  
	write(bt, "\r\n", 2);
	write(bt, "$RCMD402*2E\r\n", 13);
	write(bt, cmd1, sizeof(cmd1));
}

#define BAR_WIDTH  40

// print download progress bar
void draw_bar(int sent, int total) {
	int i, percent = (100*sent)/total;
	int rate = 0;

	printf("[");
	for(i=0;i<BAR_WIDTH;i++) {
		if(i<(BAR_WIDTH*percent/100)) putchar('#');
		else putchar('-');
	}
	printf("] %d/%d", sent, total);
	printf("    \r"); // return cursor to beginning of line
	fflush(stdout);
}

unsigned char nmea_checksum( char *str) {
// Loop through all chars to get a checksum
	unsigned char checksum = 0;
	while(*str) {	
		if(*str != '$') checksum ^= *str;
		str++;
	}
	return checksum;
}

// ask retransmit from the position
void SendLogPosition(int pos)
{
static unsigned char cmd[]={CMD_GETLOG,
		0x00,0x00,0x00,0x00,
		0x01,0x00,0x14,0x00,
		0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00
};
	cmd[9] = pos; 
	cmd[10] = pos>>8;
	send_cmd_sequence(cmd,sizeof(cmd));
}


// reverse copy for the bigendian machines
// littleendian can use memcpy
void *rcopy(char *dst,char *src,size_t len)
{
int i;
	dst += len;
	for(i=0;i<len;i++)
		*--dst = *src++;
	return dst;
}

// which copy to use straight of reverse
typedef void *(*copy_t)(void *d, void *s, size_t len);
copy_t mcopy= (copy_t) &memcpy;

// check for endian and link the proper copy
// you can do it also by multiplying 2^8 2^16 2^24, but this 
// works fine firt clear endiands 1234 or 4321 ordered bytes
int bigendian()
{
int i=1;
	if(*(char*) &i){
		mcopy = (copy_t) &memcpy;
		return 0;
	}else{
		mcopy = (copy_t) &rcopy;
		return 1;
	}
}


// conversion for char,int, long etc from the stream
#define Gchar(p)	p=(char)*s++
#define Gshort(p)	mcopy(&p,s,sizeof(short)); s+=sizeof(short)
#define Gint(p)		mcopy(&p,s,sizeof(int)); s+=sizeof(int)
#define Glong(p)	mcopy(&p,s,sizeof(long)); s+=sizeof(long)

#define Gfloat(p)	mcopy(&p,s,sizeof(float)); s+=sizeof(float)
#define Gdouble(p)	mcopy(&p,s,sizeof(double)); s+=sizeof(double)
#define Gstr(p)		strncpy(p,(char*)s,sizeof(p));s+=sizeof(p)

// routines to read and decoding frames
unsigned char *ReadDatalogState(unsigned char *s,datalog_state_t *ds)
{
	Glong(ds->entries);
	Gfloat(ds->space);
} 

// Read and convert log entry
unsigned char *ReadDataLogEntry(unsigned char *s,datalog_entry_t *de)
{
	Gshort(de->block);

  // date:
	Gchar(de->year);         // since 2000
	Gchar (de->month);
	Gchar (de->day);

	Gchar (de->dummy0);

  // time:
	Gchar (de->hour);
	Gchar (de->minute);
	Gchar (de->dummy1);
	Gchar (de->dummy2);
	Gfloat (de->second);

  // position:
	Gdouble (de->latitude);
	Gdouble (de->longitude);
	Gfloat (de->altitude);

  // 
	Gfloat (de->PDOP);
	Gfloat (de->HDOP);

	Gfloat (de->speed);  // in kph
	Gfloat (de->direction);

	Gchar (de->satellites);
	Gchar (de->event_type);
	Gshort (de->csum);
}

// read navigation configuration 
unsigned char *ReadNavigation(unsigned char *s,nav_config_t *nav)
{
	Gchar(nav->GGA);
	Gchar(nav->GLL);
	Gchar(nav->GSA);
	Gchar(nav->GSV);
	Gchar(nav->RMC);
	Gchar(nav->VTG);
	Gshort(nav->dummy0); // always 50
	Gshort(nav->elevation_mask);  // 1/10 degrees: 0..900
	Gchar(nav->DOP_mask); // 1 = PDOP, 2 = HDOP, 3 = GDOP, 4 = Disabled
	Gchar(nav->GDOP_value);
	Gchar(nav->PDOP_value);
	Gchar(nav->HDOP_value);
	Gchar(nav->degrade); // on = 1, off = 0
	Gchar(nav->SBAS); // on = 1, off = 0
	return s;
}

//read datalog config
unsigned char *ReadDatalog(unsigned char *s, datalog_config_t *dl)
{
	Glong(dl->enabled);
	Glong(dl->overwrite);
	Glong(dl->include_altitude);
	Glong(dl->interval);
	Glong(dl->constant_distance);
	Gfloat(dl->speeding_limit);
	return s; 
}

//read status        
unsigned char *ReadStatus(unsigned char *s, status_t *st)
{
	Gstr(st->name);
	Gstr(st->version);
	Glong(st->dummy0);
	Gchar(st->password);
	Gchar(st->batt);
	Gstr(st->dummy1);
	Gstr(st->mode);
	s = ReadNavigation(s,&st->navigation);
	s = ReadDatalog(s,&st->datalog);
	return s;
}


status_t xStat,*status; 
datalog_entry_t xDatalogEntry,*entry;
datalog_state_t xDatalogState,*state;
char str[128];
int degrees[2];
float minutes[2];
char dir[2];

// NMEA formatters
void doGPRMC(FILE *file)
{
char str[100];
	if((outFormat >> oGPRMC & 1)) return;
	  // build GPRMC message
	  sprintf(str, "$GPRMC,%02u%02u%06.3f,A,%02d%07.4f,%c,%03d%07.4f,%c,"
		  "%.3f,%.3f,%02u%02u%02u,,A",
		  entry->hour, entry->minute, entry->second,
		  degrees[0], minutes[0], dir[0], degrees[1], minutes[1], dir[1],
		  KPH2KNOTS(entry->speed), entry->direction,
		  entry->day, entry->month, entry->year);
	  fprintf(file, "%s*%02X\r\n", str, nmea_checksum(str));
}

void doGPGGA(FILE *file)
{
char str[100];
	if((outFormat >> oGPGGA & 1)) return;
	  // build GPGGA message
				sprintf(str, "$GPGGA,%02u%02u%06.3f,%02d%07.4f,%c,%03d%07.4f,%c,"
					"1,%02d,%.1f,%.2f,M,,M,,",
				entry->hour, entry->minute, entry->second,
				degrees[0], minutes[0], dir[0], degrees[1], minutes[1], dir[1],
				entry->satellites, entry->HDOP, entry->altitude);
	  fprintf(file, "%s*%02X\r\n", str, nmea_checksum(str));
}

void doGPGLL(FILE *file)
{
char str[100];
	if((outFormat >> oGPGLL & 1)) return;
	  // build GPGLL message
	  sprintf(str, "$GPGLL,%02d%07.4f,%c,%03d%07.4f,%c," 
		  "%02u%02u%06.3f,A",
		  degrees[0], minutes[0], dir[0], degrees[1], minutes[1], dir[1],
		  entry->hour, entry->minute, entry->second);
	  fprintf(file, "%s*%02X\r\n", str, nmea_checksum(str));
}

void doGPVTG(FILE *file)
{
char str[100];
	if((outFormat >> oGPVTG & 1)) return;
	  // build GPVTG message
	  sprintf(str, "$GPVTG,%.3f,T,,M,%.3f,N,%.3f,K", 
		  entry->direction, KPH2KNOTS(entry->speed), entry->speed);
	  fprintf(file, "%s*%02X\r\n", str, nmea_checksum(str));
}

void doGPGSA(FILE *file)
{
char str[100];
	if((outFormat >> oGPGSA & 1)) return;
	  // build GPGSA message
	  sprintf(str, "$GPGSA,A,3,,,,,,,,,,,,,%.2f,%.2f,,", 
		  entry->PDOP, entry->HDOP); 
	  fprintf(file, "%s*%02X\r\n", str, nmea_checksum(str));

	  sprintf(str, "$GPMLU,%u,%u,%u,%u",
		entry->dummy0,entry->dummy1,entry->dummy2,entry->event_type 
		  ); 
	  fprintf(file, "%s*%02X\r\n", str, nmea_checksum(str));
}

void doTABSF(FILE *file,datalog_entry_t *entry)
{
static char headersent=0;
	if(!headersent) fprintf(file,
	"#\tyy-mm-dd\thh:mm:ss.00\tlatitude\tlongitude\taltitude\tkm/h\tdir\tsat\ttype\td0\td1\td2\n");
	fprintf(file,
	"%u\t%02u%02u%02u\t%02u%02u%06.3f\t%.6f\t%.6f\t%.6f\t%.1f\t%.2f\t%u\t%u\t%u\t%u\t%u\n",
		entry->block,
		entry->year,entry->month,entry->day,
		entry->hour,entry->minute,entry->second,
		entry->latitude,entry->longitude,entry->altitude,
		entry->speed,entry->direction,
		entry->satellites,entry->event_type,
		entry->dummy0,entry->dummy1,entry->dummy2
	);
	headersent=1;
}

// long and confusing but working parser
void NNNreply_parse(unsigned char *data, int len) 
{
	static int connected = 0;  
	unsigned char *s;

	debug(1,"Received\n%s",dump("Data",data,len));
  
	switch(data[0]) {
    case CMD_ECHO:
		if(!connected) {
			connected = 1;
			send_cmd(CMD_STATUS);
		}
		break;

    case CMD_STATUS:
		if(len-1 == sizeof(status_t)) {
			char *batt_str[] = { "high", "mid", "low" };

			status = &xStat;
			ReadStatus(data+1,status);
		
			printf("Device name: %s\n", status->name);
			printf("Firmware version: %s\n", status->version);
			printf("Battery: %s\n", batt_str[status->batt]);

			printf("Datalog %s.\n", 
				status->datalog.enabled?"enabled":"disabled");
			printf("%s when datalog full.\n",
				status->datalog.overwrite?"Overwrite":"Stop");
			printf("%s altitude information into log.\n",
				status->datalog.include_altitude?"Include":"Do not include");
			if(!status->datalog.interval)
				printf("Constant time logging disabled.\n");
			else
				printf("Constant time logging: %u day, %u hr, %u min, %u sec\n",
					status->datalog.interval/86400,
					(status->datalog.interval/3600)%24,
					(status->datalog.interval/60)%60,
					status->datalog.interval%60);
			if(!status->datalog.constant_distance)
				printf("Constant distance logging disabled.\n");
			else
				printf("Constant distance logging: %um\n", 
					status->datalog.constant_distance);

			if(status->datalog.speeding_limit == 0.0) 
				printf("Speeding limit disabled.\n");
			else
				printf("Speeding limit: %.2f km/h\n", 
					status->datalog.speeding_limit);

			if(status->password) {
				if(!password) {
					fprintf(stderr, "Password required, but none specified!\n");
					req_quit = 1;
				} else {
					unsigned char cmd_password[5] = { CMD_PASSWORD };
					memmove(cmd_password+1, password, 4);
					send_cmd_sequence(cmd_password, 5);
				}
			} else
				send_cmd(CMD_LOGSTATE); // no password required, request log state

		} else {
			fprintf(stderr, "Illegal status reply size\n");
			req_quit = 1;
		}
      break;

    case CMD_PASSWORD:
		if(data[1]) {
	// password ok, request log state
			send_cmd(CMD_LOGSTATE);
		} else {
			fprintf(stderr, "Wrong password!\n");
			req_quit = 1;
		}
		break;

    case CMD_GETLOG:
		break;

    case CMD_LOGERASE:
      // log erase is always last command
		if(data[1]) printf(" successful.\n");
		else printf(" failure.\n");
		req_quit = 1;
		break;

    case CMD_LOGSTATE:
		if(len-1 == sizeof(datalog_state_t)) {
			state = &xDatalogState;
			ReadDatalogState(data+1,state);
			printf("Datalog contains %u entries\n", state->entries);
			printf("Remaining space: %.2f%%\n", 100*state->space);

			if(LogfileName && ! statonly) {
//            <----- start -----  >
				if(!state->entries) {
					fprintf(stderr, "Nothing to download!\n");
					req_quit = 1;
				} else {
					export_file = fopen(LogfileName, "w");
					if(!export_file) {
						perror("Unable to open export file");
						req_quit = 1;
					} else {
						export_num = state->entries;
						export_expect = 0;
						SendLogPosition(export_expect);
						printf("Downloading:\n");
						records = 0;
					}
				}
			} else if(erase) {
				printf("Erasing log:");
				send_cmd(CMD_LOGERASE);
			} else 
				req_quit = 1;
		}
		break;

    case CMD_DATALGENTRY:
		if(len-1 == sizeof(datalog_entry_t)) {
			entry = &xDatalogEntry;
			ReadDataLogEntry(data+1,entry);

			if(export_num) {

				if(export_expect != entry->block) {
// printf("Unexpected block %d (%d expected)\n",entry->block, export_expect);
					SendLogPosition(export_expect);
					break;  // accept only blocks in order
				}

// send alive indicator every 100 blocks (should better be once a second)
				if((export_expect % 100) == 99)
					send_cmd(CMD_ALIVE);

// draw progress bar
				draw_bar(++export_expect, export_num);
				records++;

// build parts
				degrees[0] = abs(entry->latitude);
				minutes[0] = 60.0 * (fabs(entry->latitude) - degrees[0]); 
				dir[0] = (entry->latitude < 0)?'S':'N';
				degrees[1] = abs(entry->longitude);
				minutes[1] = 60.0 * (fabs(entry->longitude) - degrees[1]); 
				dir[1] = (entry->longitude < 0)?'W':'E';
				if(outFormat){
					doGPGGA(export_file);
					doGPRMC(export_file);
					doGPGLL(export_file);
					doGPVTG(export_file);
					doGPGSA(export_file);
				}else{
					doTABSF(export_file,entry);
				}
			}
		} else if(len-1 == sizeof(datalog_end_entry_t)) {
			datalog_end_entry_t *entry = (datalog_end_entry_t*)(data+1);
							
				if(export_expect < state->entries) {
//	Not end yet test, the last was received but we miss still some, so ask resend
					printf("Not end yet %u to %u\n",export_expect,state->entries);
					SendLogPosition(export_expect);
					break;  // accept only blocks in right order
				}

			printf("\n");

			if(erase) {
				printf("Erasing log:");
				send_cmd(CMD_LOGERASE);
			} else 
				req_quit = 1;
		} else {
			fprintf(stderr, "\nIllegal datalog entry size\n");
			req_quit = 1;
		}
		break;

    default:
      printf("unknown reply %x\n", data[0]);
      break;
      
  }
}


#define STATE_AA      0
#define STATE_55      1
#define STATE_LEN_LO  2
#define STATE_LEN_HI  3
#define STATE_PAYLOAD 4
#define STATE_SUM     5
#define STATE_33      6
#define STATE_CC      7

#define NMEA_STATE_RET     0
#define NMEA_STATE_DOLLAR  1 
#define NMEA_STATE_CODE    2

// state machine to parse protocol used by rbt-3000
void protocol_parse(unsigned char byte) {
  static unsigned char buffer[1024], sum;
  static int buffer_used;
  static int state = STATE_AA;
  static int plen;

  static unsigned char nmea_buffer[16];
  static int nmea_buffer_used;  
  static int nmea_state = NMEA_STATE_DOLLAR;

  // the custom rbt-3000 parser
  switch(state) {
    case STATE_AA:
      if(byte == 0xaa) {
		state = STATE_55;
		nmea_state = NMEA_STATE_DOLLAR;
      } else {
	// run simple nmea parser in parallel to detect if we are 
	// still in nmea mode, since nmea is pure ascii we can be sure
	// we aren't in nmea mode once the rbt3000 state parser 
	// leaves aa state

	switch(nmea_state) {
	  case NMEA_STATE_RET:
	    if((byte == 0x0d) || (byte == 0x0a)) 
	      nmea_state = NMEA_STATE_DOLLAR;  // wait for dollar
	    break;

	  case NMEA_STATE_DOLLAR:
	    if(byte == '$') {
	      // start saving command code
	      nmea_buffer[0] = '$';
	      nmea_buffer_used = 1;
	      nmea_state = NMEA_STATE_CODE;
	    } else {
	      // no dollar, but no return either: wait for return
	      if((byte != 0x0d) && (byte != 0x0a)) 
		nmea_state = NMEA_STATE_RET;
	    }
	    break;
	    
	  case NMEA_STATE_CODE:
	    // store byte if enough space
	    if((byte != 0x0a) && (byte != 0x0d))
	      if(nmea_buffer_used < (sizeof(nmea_buffer)-1)) 
			nmea_buffer[nmea_buffer_used++] = byte;
	    
	    if((byte == 0x0a) || (byte == 0x0d)) {
			nmea_state = NMEA_STATE_DOLLAR;
			nmea_buffer[nmea_buffer_used++] = 0;

			printf("Got nmea sequence: %s\n", nmea_buffer);
	    }
	    break;
		}
      }
      break;

    case STATE_55:
      if(byte == 0x55) state = STATE_LEN_LO;
      else if(byte != 0xaa) state = STATE_AA;
      break;

    case STATE_LEN_LO:
      plen = byte;
      state = STATE_LEN_HI;
      break;

    case STATE_LEN_HI:
      plen += 256*byte;
      state = STATE_PAYLOAD;
      sum = 0;
      buffer_used = 0;
      break;

    case STATE_PAYLOAD:
	  if(buffer_used>=sizeof(buffer)){
		printf("******** Message too long\n");
		state = STATE_AA;
	  }else{
		buffer[buffer_used++] = byte;
		sum ^= byte;
		plen--;
		if(plen == 0) state = STATE_SUM;
	  }
      break;
      
    case STATE_SUM:
      if(sum != byte) {
// resync to new frame
		if(byte != 0xaa) state = STATE_AA;
		else state = STATE_55;
      } else
		state = STATE_33;
      break;

    case STATE_33:
      if(byte != 0x33) {
			fprintf(stderr, "******** Postamble 0x33 failed!\n");
			if(byte != 0xaa) state = STATE_AA;
			else state = STATE_55;
      } else
			state = STATE_CC;
      break;

    case STATE_CC:
      if(byte != 0xCC) {
	fprintf(stderr, "******** Postamble 0xcc failed!\n");
	if(byte != 0xaa) state = STATE_AA;
	else             state = STATE_55;
      } else {
	state = STATE_AA;
	NNNreply_parse(buffer, buffer_used);
      }
      break;
  }
}


connectSerial(char *port)
{
struct termios term;
	bt = open(port,O_RDWR);
	if(bt<0) {
		perror("NOT Connected\n");
		return 1;
	}
	printf(" Connected\n");
	tcgetattr(bt, &term);
	cfmakeraw(&term);
	tcsetattr(bt,TCSANOW,&term);
	return 0;
}


void sigproc() {

  // already received ctrl-c? */
  if(bt) close(bt);
  printf("Control C abort\n");
  if(req_quit) exit(1);

  req_quit = 1;
}

// toggling formats
void oToggleFormat(char *t, void *x)
{
	outFormat ^= (int)x;
}

void oSetFormat(char *t, void *x)
{
	outFormat = (int)x;
}


// on any exit close serialport and export file
void myExit()
{
int i;
	if(export_file){
		fclose(export_file);
		debug(3,"%s Closed\n",LogfileName);
	}
	if(bt) {
		close(bt);
		debug(3,"%s Closed\n",SerialPort);
	}
	printf("Exit blueGPS\n");
}


int testNMEA(char *t, void *x)
{
static char bf[200];
int rd;
	printf("TEST NMEA \n");
	if(connectSerial(SerialPort)) exit(1);
	for(;!req_quit;){	
		rd = read(bt, bf, sizeof(bf));
		if(rd<0) break;
		bf[rd]=0;
		printf("%s",bf);
	}
	if(bt) close(bt);
	exit(0);
}

int main(int argc, char **argv) {
int status, i, channel;
struct tm *now;
time_t t;
int endian;

	outFormat=0;  // use tabbed by default
	


// original options  
	oAdd("-d# --port","Bluetooth serial port","/dev/tty.BlueGPS282DFD-SerialPort-1",(p_t)g_text,&SerialPort);
	oAdd("-f# --file","Logfile name for downloading","?",(p_t)g_text,&LogfileName);
	oAdd("-e# --erase","Erase after download","0",(p_t)g_bool,&erase);		
	oAdd("-p# --password","Password for the logfile","",	(p_t)g_text,&password);	
	oAdd("-s# --stat","Status only","0",	(p_t)g_bool,&statonly);	
	oAdd("-t --test","Test NMEA output","",(p_t)testNMEA,0);

// formating options
	oAdd("- --gpgga","toggle global positioning system fixed data",0,(p_t)oToggleFormat,(void*)oGPGGA);
	oAdd("- --gprmc","toggle recommended minimum specific GNSS data",0,(p_t)oToggleFormat,(void*)oGPRMC);
	oAdd("- --gpgll","toggle geographic position - latitude / longitude",0,(p_t)oToggleFormat,(void*)oGPGLL);
	oAdd("- --gpvtg","toggle course over ground and ground speed",0,(p_t)oToggleFormat,(void*)oGPVTG);
	oAdd("- --gpgsa","toggle GNSS DOP and active satellites",0,(p_t)oToggleFormat,(void*)oGPGSA);
	oAdd("- --all","all gpxxx toggles on",0,(p_t)oSetFormat,(void*)oAll);
	oAdd("- --tabbed","Output is tab separated file",0,(p_t)oToggleFormat,(void*)0);

// process command line		
	oOption(argc,argv,0);
// logfile name still open - generate one
	if(*LogfileName=='?'){ 
		if(LogfileName) free(LogfileName);
		LogfileName = malloc(80);
		time(&t);
		now = localtime(&t);
		sprintf(LogfileName,"%02u%02u%02u.log",now->tm_year+1900,now->tm_mon+1,now->tm_mday);
	}

//	define endian of this machine
	endian = bigendian();
// print some debugging, 
	printf("BlueGPS %s-endian - from %s to %s %s %0x\n",
		endian?"big":"little",SerialPort,LogfileName,
		erase?"and erase datalog":"",outFormat);

// open bluetooth rfcomm - actually any serial
	if(connectSerial(SerialPort)) return 1;

	rbt3000_init();

	while(!req_quit) {
		static int got = 0;
		struct timeval tv = { 1,100000};
		int n;

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		if(bt)  FD_SET(bt, &rfds);

		if((n=select(FD_SETSIZE, &rfds, &wfds, NULL, &tv)) < 0) 
			perror("select");
		else {
			if(!n) send_cmd(CMD_ECHO);

			if(FD_ISSET(bt, &rfds)) {
				unsigned char buffer[1024];
				int rd, i;

				if((rd = read(bt, buffer, sizeof(buffer))) < 0) {
					fprintf(stderr, "Unable to read from bt: %s(%d)\n", 
					strerror(errno), errno);
					req_quit = 1;
				} else {

	debug(2,"Raw Received\n%s",dump("Data",buffer,rd));

	  // feed received bytes into parser
					for(i=0;i<rd;i++){
						protocol_parse(buffer[i]);
					}
				}
			}
		}
	}

	if(export_file) fclose(export_file);
	if(bt) close(bt);
	bt = 0;
	export_file = 0;
	printf("%u records received\n",records);
	return 0;
}
